(*========================================================================*)
(*                                                                        *)
(*                 Compositional CompcertTSO                              *)
(*                                                                        *)
(*                 Chung-Kil Hur, MPI-SWS                                 *)
(*                 Viktor Vafeiadis, MPI-SWS                              *)
(*                 Derek Dreyer, MPI-SWS                                  *)
(*                                                                        *)
(*  Copyright 2010                                                        *)
(*                                                                        *)
(*  All rights reserved.  This file is distributed under the terms of     *)
(*  the INRIA Non-Commercial License Agreement.                           *)
(*                                                                        *)
(*========================================================================*)

Require Import Coqlib.
Require Import Libtactics.
Require Import MyTactics.

(*************************
       Finite Maps
 *************************)

Record decidable_type := mkdectype {
  dt_typ :> Type;
  dt_dec : forall a b : dt_typ, {a = b} + {a <> b}
}.

Definition FMap (A: decidable_type) (B: Type) := list (dt_typ A * B).

Fixpoint FMfind {A:decidable_type} {B} (m: FMap A B) (i:A) : option B :=
  match m with
  | nil => None
  | (a,b):: m' => if dt_dec _ a i then Some b else FMfind m' i 
  end.

Definition FMequal {A B} (m1 m2: FMap A B) : Prop :=
  forall i, FMfind m1 i = FMfind m2 i.

Definition FMle {A B} (m1 m2: FMap A B) : Prop :=
  forall i e, FMfind m1 i = Some e -> FMfind m2 i = Some e.

Definition FMdunion {A B} (m1 m2 m: FMap A B) : Prop :=
  forall i,
    (FMfind m1 i = FMfind m i /\ FMfind m2 i = None)
    \/
    (FMfind m2 i = FMfind m i /\ FMfind m1 i = None).

Definition FMunion {A B} (m1 m2 m: FMap A B) : Prop :=
  forall i e,
    FMfind m i = Some e <-> (FMfind m1 i = Some e \/ FMfind m2 i = Some e).

Definition FMdjoint {A B} (m1 m2: FMap A B) : Prop :=
  exists m, FMdunion m1 m2 m.

Definition FMcompatible {A B} (m1 m2: FMap A B) : Prop :=
  exists m, FMunion m1 m2 m.

Definition FMempty {A B} (m: FMap A B) : Prop :=
  forall i, FMfind m i = None.

Definition FMkeys {A:decidable_type} {B} (m: FMap A B) : list A := map fst m.

Lemma fm_find_app_eq: forall {A B} (m1 m2: FMap A B) i, 
  FMfind (m1++m2) i = FMfind m1 i \/ FMfind (m1++m2) i = FMfind m2 i.
Proof.
intros A B m1; induction m1; intros; eauto.
by destruct a; simpl; destruct (dt_dec A d i); eauto.
Qed.

Lemma fm_find_app_none: forall {A B} (m1 m2: FMap A B) i,
  FMfind (m1++m2) i = None <-> FMfind m1 i = None /\ FMfind m2 i = None.
Proof.
intros A B m1; induction m1; intros; split; intros; des; eauto;
  destruct a; simpl; destruct (dt_dec A d i); simpl in *; 
  destruct (dt_dec A d i); subst; try edone; applyH; eauto.
Qed.

Lemma fm_find_app_iff: forall A B (m1 m2: FMap A B) i e,
  FMfind (m1 ++ m2) i = Some e <-> 
  FMfind m1 i = Some e \/ (FMfind m1 i = None /\ FMfind m2 i = Some e).
Proof.
intros A B m1; induction m1; intros.
  by split; intros; eauto; destruct H; des.
split; intros; destruct a; simpl in *; destruct (dt_dec A d i); try (by intuition); by applyH.
Qed.

Lemma fm_dunion_union: forall {A B} {m1 m2 m: FMap A B},
  FMdunion m1 m2 m -> FMunion m1 m2 m.
Proof.
split; intros; specialize (H i).
- by rewrite <- H0; destruct H; [left | right]; exactH.
- by symmetry; destruct H0; rewrite H0 in H; destruct H; 
    try (by exactH); try (by apply proj2 in H; inversion H).
Qed.  

Lemma fm_dunion_comm: forall A B (m1 m2 m: FMap A B),
  FMdunion m1 m2 m -> FMdunion m2 m1 m.
Proof. red; intros; specialize (H i); intuition. Qed.

Lemma fm_dunion_assoc: forall A B (m1 m2 m3 m12 m123: FMap A B),
  FMdunion m1 m2 m12 -> FMdunion m12 m3 m123 ->
  exists m23, FMdunion m2 m3 m23 /\ FMdunion m1 m23 m123.
Proof. 
intros; eexists (m2++m3); split; intros i; specialize (H i); specialize (H0 i);
  assert (Htmp := fm_find_app_eq m2 m3 i).

  des; try by intuition.
  - by left; intuition; rewrite H3 in H1; eapply fm_find_app_none in H1; des; auto.
  - by left; split; auto; rewrite H, H2 in H1; eapply fm_find_app_none in H1; des; auto.
  - by right; split; auto; rewrite H2 in H1; eapply fm_find_app_none in H1; des; auto.
  - by right; split; auto; rewrite H.

des; try (by left; repeat rewriteH); try (by right; repeat rewriteH).
- right; repeat rewriteH ; split; eauto.
  by rewrite <- H0; rewrite H3 in H1; eapply fm_find_app_none in H1; des; eauto.
- right; split; eauto; rewrite H1, H, H2, <- H0; rewrite H, H2 in H1.
  by eapply fm_find_app_none in H1; des; eauto.
right; split; eauto; rewrite H1, H2, <- H0, <- H; rewrite H2 in H1.
by eapply fm_find_app_none in H1; des; eauto.
Qed.

Lemma fm_union_empty: forall A B (m1 m2 m: FMap A B),
  FMunion m1 m2 m -> FMempty m1 -> FMequal m2 m.
Proof.
intros; intro i; apply option_eq; intro e; destruct (H i e); specialize (H0 i); intuition.
by rewrite H0 in H1; inversion H1.
Qed.

Lemma fm_union_comm: forall A B (m1 m2 m: FMap A B),
  FMunion m1 m2 m -> FMunion m2 m1 m.
Proof. split; specialize (H i e); intuition. Qed.

Lemma fm_union_gen: forall A B (m1 m2 m: FMap A B),
  FMunion m1 m2 m -> 
  forall i e, 
    FMfind m i = Some e ->
  (FMfind m1 i = Some e /\ FMfind m2 i = Some e) \/
  (FMfind m1 i = Some e /\ FMfind m2 i = None) \/
  (FMfind m1 i = None /\ FMfind m2 i = Some e).
Proof.
intros; specialize (H i); assert (H' := H e).
destruct (FMfind m1 i); [|by right; right; split; auto; eby intuition].
destruct (FMfind m2 i); [|by right; left; split; auto; eby intuition].
left; destruct (proj1 H' H0); inversion H1; subst; split; auto.
  by exploit (proj2 (H b0)); [right | intros; rewrite <-H0, <- H2].
by exploit (proj2 (H b)); [left | intros; rewrite <-H0, <- H2].
Qed.

Lemma fm_union_assoc: forall A B (m1 m2 m3 m12 m123: FMap A B),
  FMunion m1 m2 m12 -> FMunion m12 m3 m123 ->
  exists m23, FMunion m2 m3 m23 /\ FMunion m1 m23 m123.
Proof. 
intros; eexists (m2++m3).
assert (H' := fm_union_gen _ _ _ _ _ H).
assert (H0' := fm_union_gen _ _ _ _ _ H0).
split; red; intros. 
  split; intros.
    by eapply fm_find_app_iff in H1; intuition.
  eapply fm_find_app_iff; intuition.
  exploit (proj2 (H0 i e)); [by right|intros].
  destruct (H0' _ _ H1) as [?|[?|?]];
    try by des; destruct (H' _ _ H3); des; auto.
  des; right; split; auto.
  specialize (H i); rewrite H3 in H.
  destruct (FMfind m2 i); auto.
  by symmetry; applyH.
split; intros.
  destruct (H0' _ _ H1) as [?|[?|?]];
    try by des; destruct (H' _ _ H2) as [?|[?|?]]; intuition;
           right; eapply fm_find_app_iff; intuition.
  des; right; eapply fm_find_app_iff; right; split; auto.
  specialize (H i); rewrite H2 in H.
  destruct (FMfind m2 i); auto.
  by symmetry; applyH.
eapply H0; destruct H1; [by left; applyH|]. 
by destruct (fm_find_app_eq m2 m3 i); rewrite H1 in H2; [left; applyH | eauto].
Qed.

Lemma fm_union_equal: forall A B (m1 m1' m2 m2' m m': FMap A B),
  FMunion m1 m2 m -> FMunion m1' m2' m' ->
  FMequal m1 m1' -> FMequal m2 m2' -> FMequal m m'.
Proof.
red; intros; apply option_eq; intros.
assert (H1' := proj2 (@option_eq _ _ _) (H1 i) c).
assert (H2' := proj2 (@option_eq _ _ _) (H2 i) c).
specialize (H i c).
specialize (H0 i c).
by intuition.
Qed.

Lemma fm_union_le: forall A B (m1 m1' m2 m2' m m': FMap A B),
  FMunion m1 m2 m -> FMunion m1' m2' m' ->
  FMle m1 m1' -> FMle m2 m2' -> FMle m m'.
Proof.
by red; intros; applyH; eapply H in H3; intuition.
Qed.

Lemma fm_equal_sym: forall A B (m1 m2: FMap A B),
  FMequal m1 m2 -> FMequal m2 m1.
Proof. by intros; intro i; intuition. Qed.

Lemma fm_equal_trans: forall A B (m1 m2 m3: FMap A B),
  FMequal m1 m2 -> FMequal m2 m3 -> FMequal m1 m3.
Proof. by intros; intro i; etransitivity; eauto. Qed.

Lemma fm_keys_find: forall A B (m: FMap A B) a,
  FMfind m a <> None <-> In a (FMkeys m).
Proof.
intros; induction m.
  by intuition.
destruct a0 as (a' & b'); simpl; destruct (dt_dec _ a' a).
  eby subst; split; intros; [left|].
by intuition.
Qed.
