(*========================================================================*)
(*                                                                        *)
(*                 Compositional CompcertTSO                              *)
(*                                                                        *)
(*                 Chung-Kil Hur, MPI-SWS                                 *)
(*                 Viktor Vafeiadis, MPI-SWS                              *)
(*                 Derek Dreyer, MPI-SWS                                  *)
(*                                                                        *)
(*  Copyright 2010                                                        *)
(*                                                                        *)
(*  All rights reserved.  This file is distributed under the terms of     *)
(*  the INRIA Non-Commercial License Agreement.                           *)
(*                                                                        *)
(*========================================================================*)

Require Import Syntax.
Require Import List.
Require Import Relations.
Require Import Wellfounded.

Require Import Coqlib.
Require Import Libtactics.
Require Import MyTactics.
Require Import FiniteMap.

Set Implicit Arguments.

(**************************
   Language Specification  
 **************************)

(* Note: function body is in local memory. *)

Module Type LangBaseSig.

  Parameter thread_id : Type.
  Parameter funptr : Type.
  Parameter gtyp : Type.
  Parameter gval : Type.
  Parameter sysevent : Type.

  Parameter gval_match: gtyp -> gval -> Prop.

  Parameter thread_id_dec : forall tid tid' : thread_id, {tid = tid'} + {tid <> tid'}.
  Parameter funptr_dec : forall f f' : funptr, {f = f'} + {f <> f'}.
End LangBaseSig.

Module LanguageSpecification (LBase : LangBaseSig).
  Module Export LBaseInst := LBase.

  Definition tidtyp := mkdectype thread_id thread_id_dec.
  Definition fptrtyp := mkdectype funptr funptr_dec.

  Record gsignature : Type := mkgsignature
    { gsig_args : list gtyp;  gsig_res : option gtyp }.

  Fixpoint sigargs_match (argtyps: list gtyp) (args: list gval) : Prop :=
    match argtyps, args with
    | [], [] => True
    | typ::typtl, v::vtl => gval_match typ v /\ sigargs_match typtl vtl
    | _,_ => False
    end.

  Definition sigres_match (otyp: option gtyp) (v: gval) : Prop :=
    option_default True (fun typ => gval_match typ v) otyp.

  Inductive tausys_event :=
  | Etau
  | Eevt (sevt: sysevent)
  .

  Inductive normal_event := 
  | Esysevt (sevt: tausys_event)
  | Etstart (tid: tidtyp)
  | Etexit
  .

  Inductive event  : Type :=
  | ObsEvt (nevt: normal_event)
  | Efail
  | Eoutofmemory
  .

  Coercion Esysevt : tausys_event >-> normal_event.
  Coercion ObsEvt : normal_event >-> event.

  Definition e2ne (evt: event) : normal_event :=
    match evt with | ObsEvt ev => ev | _ => Etau end.

  Definition evt_get_tidnew (evt: event) : option tidtyp :=
    match evt with
    | Etstart tidnew => Some tidnew
    | _ => None
    end.

  Definition funset := FMap fptrtyp gsignature.

  (** * Language semantics *)

  Structure LangSem := mkLangSem {

    (* Types *)

    LS_prog: Type;
    LS_mem: Type;
    LS_tcont: Type;
    LS_tstate: Type;
    LS_state: Type  (**r contains [LS_mem], [tid -fin-> LS_tstate] *) ;

    (* Functions *)
    LS_step: tidtyp -> event -> LS_state -> LS_state -> Prop;

    LS_mefuns: LS_mem -> funset;  (**r functions exported *)
    LS_mfuns: LS_mem -> funset;   (**r all public functions (exported and imported) *)
    LS_mlocal: LS_mem -> Prop;    (**r is the memory local? *) 
    LS_mdunion: LS_mem -> LS_mem -> LS_mem -> Prop;   (**r [mdunion m1 m2 (m1 + m2)] *)

    LS_st_mem: LS_state -> LS_mem;
    LS_st_tss: LS_state -> FMap tidtyp LS_tstate;

    LS_funapp: tidtyp -> fptrtyp -> gsignature -> list gval -> LS_tcont -> 
      LS_mem -> LS_tstate -> Prop;
    LS_funret: tidtyp -> option gtyp -> gval -> LS_tcont -> 
      LS_mem -> LS_tstate -> Prop;

    LS_init: LS_prog -> 
             LS_mem -> LS_mem ->
             option (tidtyp * fptrtyp * gsignature * list gval * LS_tcont * LS_tstate) 
             -> Prop;

    (* For intermediate languages *)

    (** Intermediate languages contain normal states and transition states *)

    LS_is_intermediate : bool;
    LS_ts_normal: LS_tstate -> Prop (**r is a normal state (not special "transition" state) *);
    LS_tc_internal: LS_tcont -> Prop (**r return somewhere in the module *) ;
    LS_tc_external: LS_tcont -> Prop (**r return to another module *);
    LS_mem_external: LS_mem -> Prop (**r memory possibly used by other modules *);

    (* Properties *)

    LS_step_before: forall {st tid st' evt} (Hstep: LS_step tid evt st st'),
      option_default False LS_ts_normal (FMfind (LS_st_tss st) tid); 
    LS_step_after: forall {st tid st' evt} (Hstep: LS_step tid evt st st'),      
        evt <> Efail -> evt <> Eoutofmemory -> evt <> Etexit -> 
      FMfind (LS_st_tss st') tid <> None;
    LS_step_after_texit: forall {st tid st' evt} (Hstep: LS_step tid evt st st'),
      evt = Etexit -> FMfind (LS_st_tss st') tid = None;
    LS_step_after_tcreate: forall {st tid st' evt} (Hstep: LS_step tid evt st st'),
      forall tidnew, evt = Etstart tidnew -> 
        FMfind (LS_st_tss st) tidnew = None /\ FMfind (LS_st_tss st') tidnew <> None;
    LS_step_others: forall {st tid st' evt} (Hstep: LS_step tid evt st st'), 
      forall tid', tid' <> tid -> Some tid' <> evt_get_tidnew evt -> 
        FMfind (LS_st_tss st') tid' = FMfind (LS_st_tss st) tid';

    LS_funapp_match: forall tid f sig args ret m ts,
      LS_funapp tid f sig args ret m ts -> 
      sigargs_match sig.(gsig_args) args;

    LS_funret_match: forall tid rty v rtc m ts,
      LS_funret tid rty v rtc m ts ->
      sigres_match rty v;

    LS_funapp_monotone: forall tid f sig args ret m em m' ts,
        LS_mdunion m em m' ->
        LS_funapp tid f sig args ret m ts ->
      LS_funapp tid f sig args ret m' ts;

    LS_funret_monotone: forall tid rty v rtc m em m' ts,
        LS_mdunion m em m' ->
        LS_funret tid rty v rtc m ts ->
      LS_funret tid rty v rtc m' ts;

    (** Note: disjoint union for efuns, normal union for funs *)
   
    LS_mdunion_efuns: forall {m1 m2 m}, 
      LS_mdunion m1 m2 m -> 
      FMdunion (LS_mefuns m1) (LS_mefuns m2) (LS_mefuns m);

    LS_mdunion_funs: forall {m1 m2 m}, 
      LS_mdunion m1 m2 m -> 
      FMunion (LS_mfuns m1) (LS_mfuns m2) (LS_mfuns m);

    (** Exported functions are included in the set of public functions *)
    LS_funs_include_efuns: forall m,
      FMle (LS_mefuns m) (LS_mfuns m);

    LS_mdunion_comm: forall {m1 m2 m}, 
      LS_mdunion m1 m2 m -> LS_mdunion m2 m1 m;

    LS_mdunion_assoc: forall {m1 m2 m3 m12 m123},
      LS_mdunion m1 m2 m12 ->
      LS_mdunion m12 m3 m123 ->
      exists m23,
        LS_mdunion m2 m3 m23 /\
        LS_mdunion m1 m23 m123
  }.

  Definition LS_mdjoint {LS} (m1 m2: LS_mem LS) : Prop :=
    exists m, LS_mdunion _ m1 m2 m.

  (** A module is closed if it does import any functions, i.e. 
      its public functions are precisely its exported ones. *)
  Definition LS_mclosed LS m : Prop :=
    FMequal (LS_mefuns LS m) (LS_mfuns LS m).

  Structure TerminalLangSem := {
    TL_LS :> LangSem;

    (* Functions *)

    TL_emp: LS_mem TL_LS  (**r empty memory & no functions *);

    TL_link: LS_prog TL_LS -> LS_prog TL_LS -> LS_prog TL_LS -> Prop
         (**r [TL_link p1 p2 (p1 + p2)] *);

    (* Properties *)

    TL_is_not_intermediate: LS_is_intermediate TL_LS = false;

    TL_ts_normal_def: forall ts, LS_ts_normal TL_LS ts;

    TL_tc_internal_def: forall tc, LS_tc_internal TL_LS tc;

    TL_tc_external_def: forall tc, LS_tc_external TL_LS tc;

    TL_mem_external_def: forall m, LS_mem_external TL_LS m <-> LS_mlocal _ m;

    TL_step_progress: forall st tid, 
        FMfind (LS_st_tss _ st) tid <> None ->
      exists st' evt, LS_step TL_LS tid evt st st';

    TL_emp_empty: forall m, LS_mdunion TL_LS m TL_emp m;

    TL_emp_local: LS_mlocal TL_LS TL_emp;

    TL_emp_funs: FMempty (LS_mfuns _ TL_emp);

    (** Union of local modules/memories is local *)
    TL_mdunion_local: forall {m1 m2 m},
      LS_mdunion TL_LS m1 m2 m -> LS_mlocal _ m1 -> LS_mlocal _ m2 
      -> LS_mlocal _ m;

    TL_init_condition: forall p gm lm ots,
      LS_init TL_LS p gm lm ots
      -> LS_mlocal TL_LS lm /\ 
         LS_mdjoint gm lm /\
         match ots with
         | None => True
         | Some (tid,f,sig,args,rtc,ts) =>
           FMfind (LS_mefuns _ lm) f = Some sig /\
           LS_funapp _ tid f sig args rtc gm ts
         end;

    TL_link_decompose: forall p1 p2 p gm lm ots, 
      TL_link p1 p2 p
      -> (LS_init TL_LS p gm lm ots <->
          exists lm1 lm2 ots1 ots2, 
          LS_mdunion _ lm1 lm2 lm /\ 
          ((ots1 = ots /\ ots2 = None) \/ (ots1 = None /\ ots2 = ots)) /\
          LS_init _ p1 gm lm1 ots1 /\ 
          LS_init _ p2 gm lm2 ots2)
  }.

  (** ** Intermediate languages *)

  Structure InterLangSem :=  mkInterLangSem {
    IL_LS :> LangSem;

    (* Functions *)

    IL_mext: forall (efs fs: funset), FMle efs fs -> LS_mem IL_LS;

    IL_tcnone: LS_tcont IL_LS  (** special module-transition continuation *);

    (* Properties *)

    IL_is_intermediate: LS_is_intermediate IL_LS = true;

    IL_tc_internal_def: forall tc, LS_tc_internal IL_LS tc <-> tc <> IL_tcnone;

    IL_tc_external_def: forall tc, LS_tc_external IL_LS tc <-> tc = IL_tcnone;

    IL_mem_external_def: forall m, 
      LS_mem_external IL_LS m <-> exists efs fs Hle, m = @IL_mext efs fs Hle;

    IL_mext_local: forall efs fs (Hle: FMle efs fs), LS_mlocal _ (IL_mext Hle);
    IL_mext_efuns: forall efs fs (Hle: FMle efs fs), FMequal (LS_mefuns _ (IL_mext Hle)) efs;
    IL_mext_funs: forall efs fs (Hle: FMle efs fs), FMequal (LS_mfuns _ (IL_mext Hle)) fs;
    IL_mext_empty: forall efs fs (Hle: FMle efs fs) m, 
        FMdjoint (LS_mefuns _ m) efs -> 
        FMcompatible (LS_mfuns _ m) fs ->
      LS_mdjoint m (IL_mext Hle);
    IL_mext_identity: forall efs fs (Hle: FMle efs fs) m1 m2 m,
        LS_mdunion _ m1 (IL_mext Hle) m -> 
        LS_mdunion _ m2 (IL_mext Hle) m -> 
      m1 = m2;
    (** External calls and returns go to transition states *)
    IL_mext_transition: forall efs fs (Hle: FMle efs fs) tid lm m ts f sig args rtc, 
         FMfind efs f = Some sig ->
         LS_mdunion _ lm (IL_mext Hle) m ->
         LS_funapp IL_LS tid f sig args rtc m ts -> 
       ~LS_ts_normal _ ts /\ 
       forall rty v, ~LS_funret _ tid rty v IL_tcnone m ts;

    IL_keep_retn_excl: forall tid m ts rty v, 
      LS_funret _ tid rty v IL_tcnone m ts -> ~LS_ts_normal _ ts;

    IL_state_make:
      forall (tids : list tidtyp) (tid: tidtyp) ts m (Hin: In tid tids),
      exists st, 
      LS_st_mem IL_LS st = m /\
      FMfind (LS_st_tss _ st) tid = Some ts /\
      (forall tid', FMfind (LS_st_tss _ st) tid' <> None <-> In tid' tids);

    IL_funapp_inj: forall tid m ts f sig args otc f' sig' args' otc',
        LS_funapp IL_LS tid f  sig  args  otc  m ts ->
        LS_funapp _     tid f' sig' args' otc' m ts ->
      f = f' /\ sig = sig' /\ args = args' /\ otc = otc';

    TL_funret_inj: forall tid m ts rty otc v v',
        LS_funret IL_LS tid rty v  otc m ts -> 
        LS_funret _     tid rty v' otc m ts ->
      v = v'

  }.

End LanguageSpecification.

