(*========================================================================*)
(*                                                                        *)
(*                 Compositional CompcertTSO                              *)
(*                                                                        *)
(*          Chung-Kil Hur, MPI-SWS                                        *)
(*          Viktor Vafeiadis, MPI-SWS                                     *)
(*          Derek Dreyer, MPI-SWS                                         *)
(*                                                                        *)
(*  Copyright 2010                                                        *)
(*                                                                        *)
(*  All rights reserved.  This file is distributed under the terms of     *)
(*  the INRIA Non-Commercial License Agreement.                           *)
(*                                                                        *)
(*========================================================================*)

Require Import Logic.
Require Import Syntax.
Require Import List.
Require Import Relations.
Require Import Wellfounded.
Require Import ProofIrrelevance.
Require Import Equality.
Require Import ClassicalEpsilon. (* axiom of choice *)

Require Import Coqlib.
Require Import Libtactics.
Require Import MyTactics.
Require Import FiniteMap.

Require Import LangSpec.
Require Import ProgEquivDef.

Set Implicit Arguments.

Lemma constructive_or: forall P Q, P \/ Q -> {P} + {Q}.
Proof.
by intros; destruct (excluded_middle_informative P); [left|right]; intuition.
Qed.

Module ProgramEquivalenceLemmas (LBase : LangBaseSig).
Module Export ProgDef := ProgramEquivalenceDef LBase.

Lemma ls_taus_preord : forall LS tid, preorder _ (LS_taus LS tid).
Proof. admit. Qed.

Lemma event_texit_dec: forall evt : event, {evt = Etexit} + {evt <> Etexit}.
Proof. admit. Qed.

Lemma tausteptau_ts_matched: forall LS1 LS2 tid evt st1 st2 st1' st2',
    LS_tausteptau _ tid evt st2 st2' ->
    (LS_tausteptau _ tid evt st1 st1' \/ (st1' = st1 /\ evt = Etau)) ->
    evt <> Efail -> evt <> Eoutofmemory -> 
    ts_matched st1 st2 ->
  ts_matched st1' st2' /\
  (evt <> Etexit -> 
   FMfind (LS_st_tss LS1 st1') tid <> None /\ 
   FMfind (LS_st_tss LS2 st2') tid <> None).
Proof. admit. Qed.

Lemma ls_tausteptau_step: forall LS tid evt st st',
  LS_step LS tid evt st st' -> LS_tausteptau LS tid evt st st'.
Proof. admit. Qed.

Lemma ls_tausteptau_tstate: forall {LS tid st st'} {evt: event},
  LS_tausteptau LS tid evt st st' ->
  (option_default False (LS_ts_normal _) (FMfind (LS_st_tss _ st) tid)) /\
  (forall tidnew, evt = Etstart tidnew -> 
    FMfind (LS_st_tss _ st) tidnew = None /\ 
    FMfind (LS_st_tss _ st') tidnew <> None) /\
  (evt <> Efail -> evt <> Eoutofmemory -> evt <> Etexit -> 
    FMfind (LS_st_tss _ st') tid <> None) /\
  (evt = Etexit -> FMfind (LS_st_tss _ st') tid = None) /\
  (forall tid', tid' <> tid -> Some tid' <> evt_get_tidnew evt -> 
    FMfind (LS_st_tss _ st') tid' = FMfind (LS_st_tss _ st) tid').
Proof. admit. Qed.

Lemma dmodo_to_dmod: forall LS (dmo: DModO LS),
  dmo.(dmtso) <> None -> exists dm: DMod _, dmo = dm.
Proof. admit. Qed.

Lemma tl_emp_efuns: forall (LS: TerminalLangSem),
  FMempty (LS_mefuns _ (TL_emp LS)).
Proof. admit. Qed.

End ProgramEquivalenceLemmas.

(***************** 
   Adequacy 
 *****************)

Module ProgramEquivalenceAdequate (LBase : LangBaseSig).
Module Export ProgLemmas := ProgramEquivalenceLemmas LBase.

Definition evt_pred := tidtyp -> normal_event -> Prop.

CoInductive Trace (LS: LangSem) : LS_state LS -> Type :=
| tr_fail : forall st tid 
      (Hfail: exists st', LS_step LS tid Efail st st'),
    Trace LS st
| tr_terminate : forall st 
      (Hempty: FMempty (LS_st_tss _ st)),
    Trace LS st
| tr_step : forall st st' tid (evt: normal_event) 
      (Hstep: LS_step LS tid evt st st')
      (Htail: Trace LS st'),
    Trace LS st
.
Implicit Arguments tr_terminate [LS].
Implicit Arguments tr_fail [LS].
Implicit Arguments tr_step [LS].

Definition trace_pred LS := forall st, Trace LS st -> Prop.

Inductive trace_prefixed LS (evtset: evt_pred) (traceset: trace_pred LS)
  : forall st, Trace LS st -> Prop :=
| prefix_base : forall st trace
      (Hinst: traceset st trace),
    trace_prefixed evtset traceset trace
| prefix_step : forall st st' tid evt Hstep trtail
      (Hinevt: evtset tid evt)
      (Htail: trace_prefixed evtset traceset trtail),
    trace_prefixed evtset traceset (tr_step st st' tid evt Hstep trtail)
.

Definition fords_eq_except {wfo: WfOrd} (tids: list tidtyp) (ords ords' : FMap tidtyp wfo) :=
  forall tid', ~ In tid' tids -> FMfind ords tid' = FMfind ords' tid'.

Definition fords_tid_lt {wfo: WfOrd} tid (ords ords' : FMap tidtyp wfo) :=
  option_default False (fun ord => 
    option_default False (fun ord' => wfo.(wfo_lt) ord ord') 
      (FMfind ords' tid)) (FMfind ords tid) /\
  fords_eq_except [tid] ords ords'.

Definition fords_lt {wfo: WfOrd} (ords ords' : FMap tidtyp wfo) :=
  exists tid, fords_tid_lt tid ords ords'.

Program Definition WOunit : WfOrd :=
  @mkWfOrd unit (fun _ _ => False) _ _ .
Next Obligation. Proof. by intuition. Qed.
Next Obligation. Proof. by constructor; intuition. Qed.

Definition WOsymprod (wfo1 wfo2: WfOrd) : WfOrd :=
  mkWfOrd (t_trans _ _) 
    (wf_clos_trans _ _ (wf_symprod _ _ _ _ wfo1.(wfo_wf) wfo2.(wfo_wf))).
  
Fixpoint fords_wsl_ord (wfo: WfOrd) (tids: list tidtyp) :=
  match tids with
  | [] => WOunit
  | tid :: tl => WOsymprod wfo (fords_wsl_ord wfo tl)
  end.

Definition tau_event tidcrr (tid: tidtyp) (evt: normal_event) := 
  tid = tidcrr /\ evt = Etau.

Definition trace_equal LS st0 (trace0: Trace LS st0) :=
  fun st (trace: Trace LS st) => existT _ st trace = existT _ st0 trace0.

Definition trace_steptau LS tid evt traceset st (trace: Trace LS st) :=
    match trace with
    | tr_step _ _ tid' evt' _ trtail => 
      tid' = tid /\ evt' = evt /\
      trace_prefixed (tau_event tid) traceset trtail
    | _ => False
    end.

Definition trace_tausteptau LS tid evt (traceset: trace_pred LS) st (trace: Trace LS st) :=
  trace_prefixed (tau_event tid) (trace_steptau tid evt traceset) trace.

Definition trace_tausteptau_prefix LS tid evt 
  st0 (trace0: Trace LS st0) st (trace: Trace LS st) :=
  trace_tausteptau tid evt (trace_equal trace0) trace.

Definition trace_fails_tid LS tid := 
  fun st (trace: Trace LS st) => exists Hstep, trace = tr_fail st tid Hstep.

Definition trace_tau_fails {LS} tid st (trace: Trace LS st) :=
  trace_prefixed (tau_event tid) (trace_fails_tid tid) trace.

Definition trace_tausteptau_fails {LS} tid evt st (trace: Trace LS st) :=
  trace_tausteptau tid evt (trace_fails_tid tid) trace.

Definition executed_tids tid (evt: normal_event) : list tidtyp :=
  tid ::
  match evt_get_tidnew evt with
  | None => []
  | Some tid' => [tid']
  end.

Inductive bstep (LS: LangSem) (wfo: WfOrd) : 
  FMap tidtyp wfo * LS_state LS -> option (FMap tidtyp wfo * LS_state LS) -> Type :=
| bst_fail : forall ords st tid 
      (Hfail: LS_fails _ tid Etau st),
    bstep LS wfo (ords,st) None
| bst_terminate : forall ords st 
      (Hempty: FMempty (LS_st_tss _ st)),
    bstep LS wfo (ords,st) None
| bst_step : forall ords ords' st st' tid (evt: normal_event) 
      (Hstep: LS_tausteptau LS tid evt st st')
      (Hord: fords_eq_except (executed_tids tid evt) ords' ords),
    bstep LS wfo (ords,st) (Some (ords',st'))
| bst_stutter : forall ords ords' st
      (Hord: fords_lt ords' ords),
    bstep LS wfo (ords,st) (Some (ords',st)).

CoInductive BTrace (LS: LangSem) (wfo: WfOrd) : FMap tidtyp wfo -> LS_state LS -> Type :=
| btr_end: forall ords st (Hstep: bstep LS wfo (ords,st) None), BTrace _ wfo ords st
| btr_step: forall ords ords' st st' 
    (Hstep: bstep LS wfo (ords,st) (Some (ords',st'))) 
    (trtail: BTrace LS wfo ords' st'),
  BTrace LS wfo ords st.

Definition btr_def LS wfo ords st (btr: BTrace LS wfo ords st) : BTrace LS wfo ords st :=
  match btr with
  | btr_end ords st Hstep => btr_end Hstep
  | btr_step ords ords' st st' Hstep trtail => btr_step Hstep trtail
  end.

Lemma btr_unfold: forall LS wfo ords st (btr: BTrace LS wfo ords st), btr = btr_def btr.
Proof. by intros; case btr. Defined.

Structure simul_data (LS1 LS2: TerminalLangSem) (W: WorldStrL LS1 LS2) (R: tsrel W)
:= mksimul_data {
  
  sd_w : W;
  sd_dmm1: DModM LS1; 
  sd_dmm2: DModM LS2;
  sd_st1: LS_state LS1;
  sd_st2: LS_state LS2;
  sd_ords: FMap tidtyp (wsl_ord W);
  sd_trace2: Trace _ sd_st2;

  sd_Hmatched: ts_matched sd_st1 sd_st2;
  sd_HR : forall tid ts1 ts2 
          (Hts1: FMfind (LS_st_tss _ sd_st1) tid = Some ts1)
          (Hts2: FMfind (LS_st_tss _ sd_st2) tid = Some ts2),
        option_default False (fun ord => R tid None sd_w ord ts1 ts2) (FMfind sd_ords tid);
  sd_Hst1: dmod_mem sd_dmm1 sd_st1;
  sd_Hst2: dmod_mem sd_dmm2 sd_st2;
  sd_Hm: winv_mem _ sd_w sd_dmm1 sd_dmm2;
  sd_Hclosed2: dm_em sd_dmm2 = TL_emp _
}.

Definition get_src_bstep LS1 LS2 W R Rc (Hsimul: simulated R Rc) 
  (sd: @simul_data LS1 LS2 W R) : 
    bstep _ _ (sd_ords sd,sd_st1 sd) None 
  + ({st1' : LS_state LS1 & 
      (bstep _ _ (sd_ords sd,sd_st1 sd) (Some (sd_ords sd,st1')) 
     * bstep _ _ (sd_ords sd,st1') None)%type} 
  +  {sd' : simul_data _ _ R & 
      bstep _ _ (sd_ords sd,sd_st1 sd) (Some (sd_ords sd',sd_st1 sd'))}).

intros; destruct sd; simpl in *; destruct sd_trace3; simpl in *;
rename sd_st3 into st1; rename st into st2.

Case "fail".

apply constructive_indefinite_description in Hfail; destruct Hfail as (st' & Hfail).
assert (exists ts2, FMfind (LS_st_tss _ st2) tid = Some ts2).
  assert (Htmp := LS_step_before _ Hfail).
  by destruct (FMfind (LS_st_tss _ st2) tid); eauto.
apply constructive_indefinite_description in H; destruct H as (ts2 & Hts2).
assert (exists ts1, FMfind (LS_st_tss _ st1) tid = Some ts1).
  specialize (sd_Hmatched0 tid); rewrite Hts2 in sd_Hmatched0.
  destruct (FMfind (LS_st_tss _ st1) tid); eauto.
  eby elimtype False; apply sd_Hmatched0.
apply constructive_indefinite_description in H; destruct H as (ts1 & Hts1).

set (dm1 := mkDMod ts1 sd_dmm3).
set (dm2 := mkDMod ts2 sd_dmm4).

assert (HR := sd_HR0); specialize (HR tid _ _ Hts1 Hts2).
assert (exists ord, FMfind sd_ords0 tid = Some ord).
  eby destruct (FMfind sd_ords0 tid); eauto.
apply constructive_indefinite_description in H; destruct H as (ord & Hord).
rewrite Hord in HR; apply Hsimul in HR; destruct HR as (_&_&HR).
specialize (HR sd_w0 dm1 dm2 refl_equal refl_equal).
eapply mp; [by eapply HR; try edone; apply woth_preord|].
clear HR; intros HR; eapply mp; [by eapply (HR _ _ _ _); edone|].

intros HR'; apply constructive_or in HR'; destruct HR' as [Hfail' | HR'].
  eby left; eapply bst_fail.
by destruct HR' as (HR' & _).

Case "terminate".
left; eapply bst_terminate.
intros i; specialize (Hempty i); specialize (sd_Hmatched0 i). 
rewrite Hempty in sd_Hmatched0; destruct (FMfind (LS_st_tss _ st1) i); auto.
eby elimtype False; apply (proj1 sd_Hmatched0).

Case "step".

assert (exists ts2, FMfind (LS_st_tss _ st2) tid = Some ts2).
  assert (Htmp := LS_step_before _ Hstep).
  by destruct (FMfind (LS_st_tss _ st2) tid); eauto.
apply constructive_indefinite_description in H; destruct H as (ts2 & Hts2).
assert (exists ts1, FMfind (LS_st_tss _ st1) tid = Some ts1).
  specialize (sd_Hmatched0 tid); rewrite Hts2 in sd_Hmatched0.
  destruct (FMfind (LS_st_tss _ st1) tid); eauto.
  eby elimtype False; apply sd_Hmatched0.
apply constructive_indefinite_description in H; destruct H as (ts1 & Hts1).

set (dm1 := mkDMod ts1 sd_dmm3).
set (dm2 := mkDMod ts2 sd_dmm4).

assert (HR := sd_HR0); specialize (HR tid _ _ Hts1 Hts2).
assert (exists ord, FMfind sd_ords0 tid = Some ord).
  eby destruct (FMfind sd_ords0 tid); eauto.
apply constructive_indefinite_description in H; destruct H as (ord & Hord).
rewrite Hord in HR; apply Hsimul in HR; destruct HR as (_&_&HR).
specialize (HR sd_w0 dm1 dm2 refl_equal refl_equal).
eapply mp; [by eapply HR; try edone; apply woth_preord|].
clear HR; intros HR; eapply mp; [by eapply (HR _ _ _ _); edone|].

intros HR'; apply constructive_or in HR'; destruct HR' as [Hfail | HR'].
  apply constructive_indefinite_description in Hfail; destruct Hfail as (st1'&Hfail).
  apply constructive_indefinite_description in Hfail; destruct Hfail as (st1''&Hfail).
  destruct Hfail as (Hstep' & Hfail).
  eapply constructive_or in Hstep'; destruct Hstep' as [Hs|Hs].
    by left; eapply bst_fail; do 2 eexists; eauto using LS_taus_refl.
  right; left; eexists; split.
    by eapply bst_step; exactH.
  by eapply bst_fail; do 2 eexists; split; eauto; left; apply ls_taus_preord.

destruct HR' as (Hnotfail & HR').
do 5 (apply constructive_indefinite_description in HR'; destruct HR' as (? & HR')).
destruct HR' as ((Hstep' & Hstutter) & HR').
des; destruct H2; rename H1 into Hem.

destruct (event_texit_dec evt) as [Hevt|Hevt].
  inversion Hevt; clear Hevt; subst evt.
  destruct x1; [|eby specialize (Hstutter refl_equal); des].
  
  set (sd' := @mksimul_data _ _ _ R x0 x2 x3 x st' sd_ords0 sd_trace3).
  match goal with [_ := _ : ?P -> _ |- _] => assert (A: P) end.
    by des; eapply tausteptau_ts_matched; try (by apply Hmatched); try (by exactH);
       [eby eapply ls_tausteptau_step | left; eauto].
  set (sd'' := sd' A); unfold sd' in *; clear sd'; rename sd'' into sd'.
  match goal with [_ := _ : ?P -> _ |- _] => assert (B: P) end.
    intros; destruct (thread_id_dec tid tid0); subst.
      eby rewrite (LS_step_after_texit _ Hstep refl_equal) in Hts3.
    eapply mp; [eapply (sd_HR0 _ _ _)|].
    + eby etransitivity; [|apply Hts0]; symmetry; eapply ls_tausteptau_tstate; eauto.
    + eby etransitivity; [|apply Hts3]; symmetry; eapply LS_step_others; eauto.
    intros; destruct (FMfind sd_ords0 tid0); try edone.
    eapply Hsimul; eauto.
    by eapply woth_step; eauto using @woth_refl; red; intros Htmp; applyH; inversion Htmp.
  set (sd'' := sd' B); unfold sd' in *; clear sd'; rename sd'' into sd'.
  repeat (let X := fresh "X" in 
    match goal with [_ := _ : ?P -> _ |- _] => assert (X: P) end; 
    [by simpl in Hem; try rewriteD <- Hem; exactH|];
    set (sd'' := sd' X); unfold sd' in *; clear sd'; rename sd'' into sd').
  by right; right; eexists sd'; eapply bst_step; eauto.

eapply mp; [eapply (dmodo_to_dmod x3)|]. 
  eby rewriteD <- H; eapply LS_step_after; eauto.
intros Hdm; apply constructive_indefinite_description in Hdm.    
destruct Hdm as (dm2' & Heq); subst x3.
specialize (H4 _ refl_equal); apply constructive_indefinite_description in H4.
destruct H4 as (dm1' & Heq & H4); subst x2.
eapply constructive_or in H4; destruct H4 as [H4 | H4].
  2: eapply constructive_or in H4; destruct H4 as [H4 | H4].
    2: eby eapply constructive_indefinite_description in H4; destruct H4 as (?&?&?).
  2: repeat (eapply constructive_indefinite_description in H4; destruct H4 as (?&H4)).
  2: destruct H4 as (? & (?&?&?) & ?).
  2: eby simpl in Hem; rewriteD <- Hem in H5; rewrite sd_Hclosed3, tl_emp_efuns in H5.
eapply constructive_indefinite_description in H4; destruct H4 as (ord' & H4).

destruct x1.
  destruct (option_dec (evt_get_tidnew evt)) as [Hevt'|Hevt'].  
    set (ords' := (tid,ord')::sd_ords0).
    set (sd' := @mksimul_data _ _ _ R x0 dm1' dm2' x st' ords' sd_trace3).
    match goal with [_ := _ : ?P -> _ |- _] => assert (A: P) end.
      by des; eapply tausteptau_ts_matched; try (by apply Hmatched); try (by exactH);
         [eby eapply ls_tausteptau_step | left; eauto].
    set (sd'' := sd' A); unfold sd' in *; clear sd'; rename sd'' into sd'.
    match goal with [_ := _ : ?P -> _ |- _] => assert (B: P) end.
      intros; destruct (thread_id_dec tid tid0); try subst tid0.
        simpl; destruct (thread_id_dec tid tid); try edone.
        simpl; rewriteD H in Hts0; rewriteD H in Hts3; simpl in *.
        eby inversion Hts0; inversion Hts3; subst ts0 ts3; des.
      eapply mp; [eapply (sd_HR0 _ _ _)|].
      + etransitivity; [|apply Hts0]; symmetry; eapply ls_tausteptau_tstate; eauto.
        eby rewrite Hevt'.
      + etransitivity; [|apply Hts3]; symmetry; eapply LS_step_others; eauto.
        eby rewrite Hevt'.
      intros; simpl; destruct (thread_id_dec tid tid0); try edone.
      simpl in *; destruct (FMfind sd_ords0 tid0); try edone.
      eapply Hsimul; eauto.
      by eapply woth_step; eauto using @woth_refl; red; intros Htmp; applyH; inversion Htmp.
    set (sd'' := sd' B); unfold sd' in *; clear sd'; rename sd'' into sd'.
    repeat (let X := fresh "X" in 
      match goal with [_ := _ : ?P -> _ |- _] => assert (X: P) end; 
      [by simpl in Hem; try rewriteD <- Hem; exactH|];
      set (sd'' := sd' X); unfold sd' in *; clear sd'; rename sd'' into sd').
    right; right; eexists sd'; eapply bst_step; eauto.
    red; intros; simpl; destruct (thread_id_dec tid tid'); eauto.
    by subst tid'; simpl in H1; intuition.

  destruct evt; try edone; rename tid0 into tidnew.
  specialize (H3 _ refl_equal).
  repeat (eapply constructive_indefinite_description in H3; destruct H3 as (?&H3)).
  rename x3 into ordnew.
  set (ords' := (tid,ord')::(tidnew,ordnew)::sd_ords0).
  set (sd' := @mksimul_data _ _ _ R x0 dm1' dm2' x st' ords' sd_trace3).
  match goal with [_ := _ : ?P -> _ |- _] => assert (A: P) end.
    by des; eapply tausteptau_ts_matched; try (by apply Hmatched); try (by exactH);
       [eby eapply ls_tausteptau_step | left; eauto].
  set (sd'' := sd' A); unfold sd' in *; clear sd'; rename sd'' into sd'.
  match goal with [_ := _ : ?P -> _ |- _] => assert (B: P) end.
    intros; destruct (thread_id_dec tid tid0); try subst tid0.
      simpl; destruct (thread_id_dec tid tid); try edone; simpl.
      rewriteD H in Hts0; rewriteD H in Hts3; simpl in *.
      eby inversion Hts0; inversion Hts3; subst ts0 ts3; des.
    destruct (thread_id_dec tidnew tid0); try subst tid0.
      simpl; destruct (thread_id_dec tid tidnew); try edone.
      destruct (thread_id_dec tidnew tidnew); try edone.
      rewriteD H3 in Hts0; rewriteD H3 in Hts3; inversion Hts0; inversion Hts3.
      eby subst x1 x2; des.
    eapply mp; [eapply (sd_HR0 _ _ _)|].
    + etransitivity; [|apply Hts0]; symmetry; eapply ls_tausteptau_tstate; eauto.
      eby red; intros Htmp; apply n0; inversion Htmp.
    + etransitivity; [|apply Hts3]; symmetry; eapply LS_step_others; eauto.
      eby red; intros Htmp; apply n0; inversion Htmp.
    intros; simpl; destruct (thread_id_dec tid tid0); try edone.
    simpl in *; destruct (thread_id_dec tidnew tid0); try edone.
    destruct (FMfind sd_ords0 tid0); try edone.
    eapply Hsimul; eauto.
    by eapply woth_step; eauto using @woth_refl; red; intros Htmp; applyH; inversion Htmp.
  set (sd'' := sd' B); unfold sd' in *; clear sd'; rename sd'' into sd'.
  repeat (let X := fresh "X" in 
    match goal with [_ := _ : ?P -> _ |- _] => assert (X: P) end; 
    [by simpl in Hem; try rewriteD <- Hem; exactH|];
    set (sd'' := sd' X); unfold sd' in *; clear sd'; rename sd'' into sd').
  right; right; eexists sd'; eapply bst_step; eauto.
  red; intros; simpl; destruct (thread_id_dec tid tid').
    by inversion e; subst tid'; simpl in H1; intuition.
  destruct (thread_id_dec tidnew tid'); eauto.
  by subst tid'; simpl in H1; intuition.

specialize (Hstutter refl_equal); des.
inversion H6; clear H6; subst evt x.
set (ords' := (tid,ord')::sd_ords0).
set (sd' := @mksimul_data _ _ _ R x0 dm1' dm2' st1 st' ords' sd_trace3).
match goal with [_ := _ : ?P -> _ |- _] => assert (A: P) end.
  by des; eapply tausteptau_ts_matched; try (by apply Hmatched); try (by exactH);
     [eapply ls_tausteptau_step; edone | right; eauto].
set (sd'' := sd' A); unfold sd' in *; clear sd'; rename sd'' into sd'.
match goal with [_ := _ : ?P -> _ |- _] => assert (B: P) end.
  intros; destruct (thread_id_dec tid tid0); try subst tid0.
    simpl; destruct (thread_id_dec tid tid); try edone.
    simpl; rewriteD H in Hts0; rewriteD H in Hts3; simpl in *.
    eby inversion Hts0; inversion Hts3; subst ts0 ts3; des.
  eapply mp; [eapply (sd_HR0 _ _ _)|].
  + exactH.
  + eby etransitivity; [|apply Hts3]; symmetry; eapply LS_step_others; eauto.
  intros; simpl; destruct (thread_id_dec tid tid0); try edone.
  simpl in *; destruct (FMfind sd_ords0 tid0); try edone.
  eapply Hsimul; eauto.
  by eapply woth_step; eauto using @woth_refl; red; intros Htmp; applyH; inversion Htmp.
set (sd'' := sd' B); unfold sd' in *; clear sd'; rename sd'' into sd'.
repeat (let X := fresh "X" in 
  match goal with [_ := _ : ?P -> _ |- _] => assert (X: P) end; 
  [by simpl in Hem; try rewriteD <- Hem; exactH|];
  set (sd'' := sd' X); unfold sd' in *; clear sd'; rename sd'' into sd').
right; right; eexists sd'; eapply bst_stutter.
eexists tid; split.
  simpl; destruct (thread_id_dec tid tid); [|edone].
  by simpl; rewrite Hord; applyH.
red; intros; simpl; destruct (thread_id_dec tid tid'); eauto.
by subst tid'; simpl in H1; intuition.
Defined.

CoFixpoint get_src_btrace LS1 LS2 W R Rc (Hsimul: simulated R Rc)
  (sd: @simul_data LS1 LS2 W R) : BTrace LS1 _ (sd_ords sd) (sd_st1 sd) 
:=
  match get_src_bstep Hsimul sd with
  | inl bs => btr_end bs
  | inr (inl (existT st1' (Hstep,Hfail))) => btr_step Hstep (btr_end Hfail)
  | inr (inr (existT sd' bs)) => btr_step bs (get_src_btrace Hsimul sd')
  end.

Ltac dest_constructive_exists :=
  match goal with [|- context[@constructive_indefinite_description ?X ?Y ?Z]] =>
    destruct (@constructive_indefinite_description X Y Z)
  end.

Ltac dest_constructive_or :=
  match goal with [|- context[@constructive_or ?X ?Y ?Z]] =>
    destruct (@constructive_or X Y Z)
  end.

Ltac dest_rel := 
  match goal with [|- context[?t ?a1 ?a2 ?a3 ?a4 ?a5 ?a6 ?a7]] => 
    destruct (t a1 a2 a3 a4 a5 a6 a7) as (?&?&?); unfold mp; simpl
  end.

Ltac generalize_proofs :=
  match goal with [|- context[@mksimul_data _ _ _ _ _ _ _ _ _ _ _ 
    ?X1 ?X2 ?X3 ?X4 ?X5 ?X6]] => generalize X1 X2 X3 X4 X5 X6
  end;
  try match goal with [|- context[@bst_step _ _ _ _ _ _ _ _ _ ?X]] => generalize X end;
  try match goal with [|- context[@bst_stutter _ _ _ _ _ _ ?X]] => generalize X end.

(****************************************************
  Proof using parameterized coinduction
  * proof construction takes 35.6 seconds
  * proof checking takes 40 seconds
 ****************************************************)

CoInductive trace_simulated_p LS1 LS2 (wfo: WfOrd) 
  (r: forall ords st1 st2, BTrace LS1 wfo ords st1 -> Trace LS2 st2 -> Prop) : 
  forall ords st1 st2, BTrace LS1 wfo ords st1 -> Trace LS2 st2 -> Prop := 

| trsimul_p_fail: forall st1 st2 ords tid (trace2: Trace _ st2)
      (Hfail: LS_fails _ tid Etau st1),
    trace_simulated_p r (btr_end (bst_fail _ ords Hfail)) trace2

| trsimul_p_terminate: forall st1 st2 ords Hempty1 Hempty2,
    trace_simulated_p r
      (btr_end (bst_terminate _ _ ords st1 Hempty1))
      (tr_terminate st2 Hempty2)

| trsimul_p_step: forall st1 st1' st2 st2' ords ords' tid evt
  trtail1 trtail2 (Hstep1: bstep LS1 wfo (ords,st1) (Some (ords',st1'))) Hstep2
      (Hmatched: ts_matched st1 st2)
      (Htail: trace_simulated_p r trtail1 trtail2 \/ r _ _ _ trtail1 trtail2),
    trace_simulated_p r
      (btr_step Hstep1 trtail1) 
      (tr_step st2 st2' tid evt Hstep2 trtail2).
Hint Constructors trace_simulated_p.

Lemma trsim_acc: forall LS1 LS2 wfo l r0
    (ACC: forall r (INC: r0 <5= r) (CIH: l <5= r), 
          l <5= @trace_simulated_p LS1 LS2 wfo r),
  l <5= trace_simulated_p r0.
Proof. 
  intros; cut (trace_simulated_p (fun a b c d e=>l a b c d e \/ r0 a b c d e) d e); eauto.
  clear PR; revert a b c d e; cofix CIH.
  intros; destruct H; [| |econstructor 3; des]; eauto 10.
Qed.

Lemma get_src_btrace_simulated_p: forall LS1 LS2 W R Rc Hsimul sd,
  trace_simulated_p (fun _ _ _ _ _ => False)
    (@get_src_btrace LS1 LS2 W R Rc Hsimul sd) (sd_trace2 sd).
Proof.
  intros; destruct Hsimul as (? & ? & ? & ?).

  (* coinduction *)
  eapply trsim_acc with (l := fun a b c d e => 
    exists sd, a~=_ sd /\ b~=_ sd /\ c~=_ sd /\ d~=_ sd /\ e~=_ sd); [|by eauto 10].
  intros; clear sd; des; subst; simpl_JMeq; subst.

  rewrite (btr_unfold (get_src_btrace _ sd)).
  destruct sd; destruct sd_trace3.

  Case "fail".

  simpl; repeat dest_constructive_exists.
  dest_rel.
  dest_constructive_or.
    by apply trsimul_p_fail.
  eby destruct a.

  Case "terminate".

  simpl; repeat dest_constructive_exists.
  by apply trsimul_p_terminate.

  Case "step".
  simpl; do 3 dest_constructive_exists.
  dest_rel.
  dest_constructive_or.
    do 2 dest_constructive_exists.
    destruct a.
    dest_constructive_or.
      by apply trsimul_p_fail.
    eapply trsimul_p_step; exactH.
    by left; apply trsimul_p_fail.
  destruct a.
  do 5 dest_constructive_exists.
  destruct t4; destruct a; destruct a0 as (? & ? & ? & ? & ?); destruct d0.
  destruct (event_texit_dec evt).
    inversion e9; subst evt. 
    rewrite (UIP_refl _ _ e9); simpl.
    unfold eq_rect_r, eq_rect, Logic.eq_sym.
    destruct x4.
      generalize_proofs. clear - CIH sd_Hmatched0; intros.
      eapply trsimul_p_step; exactH.
      by match goal with [|-context[get_src_btrace _ ?sd]] => right; apply CIH; exists sd end.
    by destruct (a eq_refl).

  dest_constructive_exists.
  unfold eq_rect_r, eq_rect, Logic.eq_sym; subst x6. 
  dest_constructive_exists.
  destruct a0; subst x5.
  dest_constructive_or.
    dest_constructive_exists.
    destruct x4.
      destruct evt; [| |edone].
        destruct (option_dec (@None thread_id)); [|edone].
        generalize_proofs; clear - CIH sd_Hmatched0; intros.
        eapply trsimul_p_step; exactH.
        by match goal with [|-context[get_src_btrace _ ?sd]] => right; apply CIH; exists sd end.
      destruct (option_dec (@Some thread_id tid0)); [edone|].
      do 3 dest_constructive_exists.
      generalize_proofs; clear - CIH sd_Hmatched0; intros.
      eapply trsimul_p_step; exactH.
      by match goal with [|-context[get_src_btrace _ ?sd]] => right; apply CIH; exists sd end.
    destruct a0; destruct (a eq_refl).
    inversion e10; subst x2 evt.
    rewrite (UIP_refl _ _ e10); simpl.
    generalize_proofs; clear - CIH sd_Hmatched0 e1 a0; intros.
    eapply trsimul_p_step; exactH.
    by right; match goal with [|-context[get_src_btrace _ ?sd]] => eapply CIH; exists sd end.
  dest_constructive_or.
    dest_constructive_exists.
    by destruct d0; edone.
  do 5 dest_constructive_exists.
  destruct d0; destruct a0; destruct c; destruct c0.
  by match goal with [|-context[False_rect _ ?t]] => contradiction t end.
Qed.

(****************************************************
  Proof using cofix
  * proof construction takes 34 seconds
  * proof checking takes 192 seconds
 ****************************************************)

CoInductive trace_simulated LS1 LS2 (wfo: WfOrd) : 
  forall ords st1 st2, BTrace LS1 wfo ords st1 -> Trace LS2 st2 -> Prop := 

| trsimul_fail: forall st1 st2 ords tid (trace2: Trace _ st2)
      (Hfail: LS_fails _ tid Etau st1),
    trace_simulated (btr_end (bst_fail _ ords Hfail)) trace2

| trsimul_terminate: forall st1 st2 ords Hempty1 Hempty2,
    trace_simulated 
      (btr_end (bst_terminate _ _ ords st1 Hempty1))
      (tr_terminate st2 Hempty2)

| trsimul_step: forall st1 st1' st2 st2' ords ords' tid evt
  trtail1 trtail2 (Hstep1: bstep LS1 wfo (ords,st1) (Some (ords',st1'))) Hstep2
      (Hmatched: ts_matched st1 st2)
      (Htail: trace_simulated trtail1 trtail2),
    trace_simulated
      (btr_step Hstep1 trtail1) 
      (tr_step st2 st2' tid evt Hstep2 trtail2).
Hint Constructors trace_simulated.

Lemma get_src_btrace_simulated: forall LS1 LS2 W R Rc Hsimul sd,
  trace_simulated (@get_src_btrace LS1 LS2 W R Rc Hsimul sd) (sd_trace2 sd).
Proof.
  intros; destruct Hsimul as (? & ? & ? & ?).

  (* coinduction *)
  revert sd; cofix CIH; intros.

  rewrite (btr_unfold (get_src_btrace _ sd)).
  destruct sd; destruct sd_trace3.

  Case "fail".

  simpl; repeat dest_constructive_exists.
  dest_rel.
  dest_constructive_or.
    by apply trsimul_fail.
  eby destruct a.

  Case "terminate".

  simpl; repeat dest_constructive_exists.
  by apply trsimul_terminate.

  Case "step".
  simpl; do 3 dest_constructive_exists.
  dest_rel.
  dest_constructive_or.
    do 2 dest_constructive_exists.
    destruct a.
    dest_constructive_or.
      by apply trsimul_fail.
    eapply trsimul_step; exactH.
    by apply trsimul_fail.
  destruct a.
  do 5 dest_constructive_exists.
  destruct t4; destruct a; destruct a0 as (? & ? & ? & ? & ?); destruct d0.
  destruct (event_texit_dec evt).
    inversion e9; subst evt. 
    rewrite (UIP_refl _ _ e9); simpl.
    unfold eq_rect_r, eq_rect, Logic.eq_sym.
    destruct x4.
      generalize_proofs; clear - CIH sd_Hmatched0; intros.
      eapply trsimul_step; exactH.
      by eapply CIH; eauto.
    by destruct (a eq_refl).

  dest_constructive_exists.
  unfold eq_rect_r, eq_rect, Logic.eq_sym; subst x6. 
  dest_constructive_exists.
  destruct a0; subst x5.
  dest_constructive_or.
    dest_constructive_exists.
    destruct x4.
      destruct evt; [| |edone].
        destruct (option_dec (@None thread_id)); [|edone].
        generalize_proofs; clear - CIH sd_Hmatched0; intros.
        eapply trsimul_step; exactH.
        by eapply CIH; eauto.
      destruct (option_dec (@Some thread_id tid0)); [edone|].
      do 3 dest_constructive_exists.
      generalize_proofs; clear - CIH sd_Hmatched0; intros.
      eapply trsimul_step; exactH.
      by eapply CIH; eauto.
    destruct a0; destruct (a eq_refl).
    inversion e10; subst x2 evt.
    rewrite (UIP_refl _ _ e10); simpl.
    generalize_proofs; clear - CIH sd_Hmatched0 e1 a0; intros.
    eapply trsimul_step; exactH.
    by match goal with [|-context[get_src_btrace _ ?sd]] => eapply (CIH sd) end.
  dest_constructive_or.
    dest_constructive_exists.
    by destruct d0; edone.
  do 5 dest_constructive_exists.
  destruct d0; destruct a0; destruct c; destruct c0.
  by match goal with [|-context[False_rect _ ?t]] => contradiction t end.
Qed.

End ProgramEquivalenceAdequate.


