Require Import Program Setoid.

Set Implicit Arguments.
Unset Strict Implicit.

(*----------------------------------------------------------------------------*)
(** * Complete lattices (based on preorders) *) 
(*----------------------------------------------------------------------------*)

Record cola := Build_cola 
{ cola_carrier :> Type
; cola_le    : cola_carrier -> cola_carrier -> Prop
; cola_lub   : (cola_carrier -> Prop) -> cola_carrier 
  (* The following operations can be defined in terms of lub.
     But for efficiency, we have them explicitly. *)  
; cola_glb   : (cola_carrier -> Prop) -> cola_carrier 
; cola_join  : cola_carrier -> cola_carrier -> cola_carrier
; cola_meet  : cola_carrier -> cola_carrier -> cola_carrier

; cola_refl  : forall a, cola_le a a 
; cola_trans : forall x y z, cola_le x y -> cola_le y z -> cola_le x z
; cola_lub_upper :
    forall (c : cola_carrier -> Prop) x, c x -> cola_le x (cola_lub c)
; cola_lub_least : 
    forall (c : cola_carrier -> Prop) x, 
    (forall n, c n -> cola_le n x) -> cola_le (cola_lub c) x
; cola_glb_lower :
    forall (c : cola_carrier -> Prop) x, c x -> cola_le (cola_glb c) x
; cola_glb_greatest : 
    forall (c : cola_carrier -> Prop) x, 
   (forall n, c n -> cola_le x n) -> cola_le x (cola_glb c)
; cola_join_upper_l : forall a b, cola_le a (cola_join a b)
; cola_join_upper_r : forall a b, cola_le b (cola_join a b)
; cola_join_least   : 
    forall a b c, cola_le a c -> cola_le b c -> cola_le (cola_join a b) c
; cola_meet_lower_l : forall a b, cola_le (cola_meet a b) a 
; cola_meet_lower_r : forall a b, cola_le (cola_meet a b) b
; cola_meet_greatest: 
    forall a b c, cola_le c a -> cola_le c b -> cola_le c (cola_meet a b)
}.
Hint Immediate cola_refl.
Hint Unfold reflexive.
Bind Scope cola_scope with cola.
Infix "<=" := cola_le : cola_scope.
Notation "x \\// y" := (cola_join x y) (at level 50, no associativity).
Notation "x //\\ y" := (cola_meet x y) (at level 50, no associativity).
Arguments cola_join c x y : simpl never.
Arguments cola_meet c x y : simpl never.

Local Open Scope cola_scope.


(** Least and greatest elements.
*)

Definition cola_bot {O : cola} := cola_glb (fun x : O => True).
Definition cola_top {O : cola} := cola_lub (fun x : O => True).

Lemma cola_bot_least : forall {O : cola} (x : O), cola_bot <= x.
Proof.
  eauto using cola_glb_lower.
Qed.

Lemma cola_top_greatest : forall {O : cola} (x : O), x <= cola_top.
Proof.
  eauto using cola_lub_upper.
Qed.


(** Properties of (binary) join and meet.
*)

Lemma cola_join_mon : forall {O : cola} {x x' y y' : O},
  x <= x' -> y <= y' -> x \\// y <= x' \\// y'.
Proof.
  intros; apply cola_join_least; eapply cola_trans; 
   eauto using cola_join_upper_l, cola_join_upper_r.  
Qed.

Lemma cola_meet_mon : forall {O : cola} {x x' y y' : O},
  x <= x' -> y <= y' -> x //\\ y <= x' //\\ y'.
Proof.
  intros; apply cola_meet_greatest; eapply cola_trans;
   [apply cola_meet_lower_l| |apply cola_meet_lower_r|]; eauto.  
Qed.


(** Canonical complete lattice for Prop.
*)

Section cola_prop.

  Definition prop_le (P Q: Prop) := forall (IN: P), Q.
  Definition prop_lub (X: Prop -> Prop) := exists P, X P /\ P. 
  Definition prop_glb (X: Prop -> Prop) := forall P, X P -> P. 
  Definition prop_join P Q := P \/ Q.
  Definition prop_meet P Q := P /\ Q.

  Program Canonical Structure prop_cola : cola := 
    @Build_cola _ prop_le prop_lub prop_glb prop_join prop_meet _ _ _ _ _ _ _ _ _ _ _ _.
  Next Obligation. repeat red; eauto. Qed.
  Next Obligation. repeat red; eauto. Qed.
  Next Obligation. repeat red; eauto. Qed.
  Next Obligation. repeat red; intros; destruct IN as (?&?&?); eapply H; eauto. Qed.
  Next Obligation. repeat intro; apply IN; eauto. Qed.
  Next Obligation. repeat intro; apply H; eauto. Qed.
  Next Obligation. repeat red; eauto. Qed.
  Next Obligation. repeat red; eauto. Qed.
  Next Obligation. repeat intro; destruct IN; eauto. Qed.
  Next Obligation. repeat intro; apply IN. Qed.
  Next Obligation. repeat intro; apply IN. Qed.
  Next Obligation. repeat intro; split; eauto. Qed.

End cola_prop.


(** Canonical complete lattice for function space (pointwise lifting).
*)

Section cola_fun.

  Variables (A:Type) (O:cola).

  Definition fun_le (f g: A->O) := forall x, f x <= g x.
  Definition fun_lub (X: (A -> O) -> Prop) := 
    fun x => cola_lub (fun y => exists f, X f /\ y = f x). 
  Definition fun_glb (X: (A -> O) -> Prop) := 
    fun x => cola_glb (fun y => exists f, X f /\ y = f x). 
  Definition fun_join (f g: A -> O) a := f a \\// g a.
  Definition fun_meet (f g: A -> O) a := f a //\\ g a.

  Program Canonical Structure fun_cola : cola :=
    @Build_cola _ fun_le fun_lub fun_glb fun_join fun_meet _ _ _ _ _ _ _ _ _ _ _ _ .
  Next Obligation. repeat intro; eauto. Qed.
  Next Obligation. repeat intro; eauto using cola_trans. Qed.
  Next Obligation. repeat intro; eauto using cola_lub_upper. Qed.
  Next Obligation. 
    intro; apply cola_lub_least; intros ? (?,(?,?)); subst; apply H; auto.
  Qed.
  Next Obligation. repeat intro; eauto using cola_glb_lower. Qed.
  Next Obligation. 
    intro; apply cola_glb_greatest; intros ? (?,(?,?)); subst; apply H; auto.
  Qed.
  Next Obligation. repeat intro; eauto using cola_join_upper_l. Qed.
  Next Obligation. repeat intro; eauto using cola_join_upper_r. Qed.
  Next Obligation. repeat intro; eauto using cola_join_least. Qed.
  Next Obligation. repeat intro; eauto using cola_meet_lower_l. Qed.
  Next Obligation. repeat intro; eauto using cola_meet_lower_r. Qed.
  Next Obligation. repeat intro; eauto using cola_meet_greatest. Qed.

End cola_fun.


(** Canonical complete lattice for dependent function space (pointwise
    lifting).

    Unfortunately, the canonical structure here doesn't really work,
    so one has to be a bit more explicit when working with a dependent
    function space.
*)

Section cola_depfun.

  Variables (A:Type) (O:A -> cola).

  Definition depfun_le (f g : forall a : A, O a) := forall x, f x <= g x.
  Definition depfun_lub (X: (forall a : A, O a) -> Prop) := 
    fun x => cola_lub (fun y => exists f, X f /\ y = f x). 
  Definition depfun_glb (X: (forall a : A, O a) -> Prop) := 
    fun x => cola_glb (fun y => exists f, X f /\ y = f x). 
  Definition depfun_join (f g: forall a : A, O a) a := f a \\// g a.
  Definition depfun_meet (f g: forall a : A, O a) a := f a //\\ g a.

  Program Canonical Structure depfun_cola : cola :=
    @Build_cola _ depfun_le depfun_lub depfun_glb depfun_join depfun_meet 
      _ _ _ _ _ _ _ _ _ _ _ _.
  Next Obligation. repeat intro; eauto. Qed.
  Next Obligation. repeat intro; eauto using cola_trans. Qed.
  Next Obligation. repeat intro; eauto using cola_lub_upper. Qed.
  Next Obligation. 
    intro; apply cola_lub_least; intros ? (?,(?,?)); subst; apply H; auto.
  Qed.
  Next Obligation. repeat intro; eauto using cola_glb_lower. Qed.
  Next Obligation. 
    intro; apply cola_glb_greatest; intros ? (?,(?,?)); subst; apply H; auto.
  Qed.
  Next Obligation. repeat intro; eauto using cola_join_upper_l. Qed.
  Next Obligation. repeat intro; eauto using cola_join_upper_r. Qed.
  Next Obligation. repeat intro; eauto using cola_join_least. Qed.
  Next Obligation. repeat intro; eauto using cola_meet_lower_l. Qed.
  Next Obligation. repeat intro; eauto using cola_meet_lower_r. Qed.
  Next Obligation. repeat intro; eauto using cola_meet_greatest. Qed.

End cola_depfun.


(** Canonical complete lattice for products.
*)

Section cola_product.

  Variables A B: cola.

  Definition prod_le (p q : A*B) := fst p <= fst q /\ snd p <= snd q.
  Definition prod_lub (X: A*B -> Prop) := 
    (cola_lub (fun a => exists b, X (a,b)), cola_lub (fun b => exists a, X (a,b))).
  Definition prod_glb (X: A*B -> Prop) := 
    (cola_glb (fun a => exists b, X (a,b)), cola_glb (fun b => exists a, X (a,b))).
  Definition prod_join (p q: A*B) := (fst p \\// fst q, snd p \\// snd q).
  Definition prod_meet (p q: A*B) := (fst p //\\ fst q, snd p //\\ snd q).

  Program Canonical Structure prod_cola : cola :=
    @Build_cola _ prod_le prod_lub prod_glb prod_join prod_meet _ _ _ _ _ _ _ _ _ _ _ _.
  Next Obligation. split; eauto. Qed.
  Next Obligation. red; destruct H, H0; eauto using cola_trans. Qed.
  Next Obligation. split; simpl; eauto using cola_lub_upper. Qed.
  Next Obligation. 
    split; apply cola_lub_least; intros ? (?,?); apply H in H0; apply H0. 
  Qed.
  Next Obligation. split; simpl; eauto using cola_glb_lower. Qed.
  Next Obligation.
    split; apply cola_glb_greatest; intros ? (?,?); apply H in H0; apply H0. 
  Qed.
  Next Obligation. split; simpl; eauto using cola_join_upper_l. Qed.
  Next Obligation. split; simpl; eauto using cola_join_upper_r. Qed.
  Next Obligation. red; destruct H, H0; simpl in *; eauto using cola_join_least. Qed.
  Next Obligation. split; simpl; eauto using cola_meet_lower_l. Qed.
  Next Obligation. split; simpl; eauto using cola_meet_lower_r. Qed.
  Next Obligation. red; destruct H, H0; simpl in *; eauto using cola_meet_greatest. Qed.

End cola_product.


(** Canonical dual complete lattice.
*)

Section cola_dual.

  Variables A: cola.

  Definition dual_le := fun p q : A => q <= p.
  Definition dual_lub (X: A -> Prop) := cola_glb X.
  Definition dual_glb (X: A -> Prop) := cola_lub X.
  Definition dual_join (a b: A) := cola_meet a b.
  Definition dual_meet (a b: A) := cola_join a b.

  Program Canonical Structure dual_cola : cola :=
    @Build_cola _ dual_le dual_lub dual_glb dual_join dual_meet _ _ _ _ _ _ _ _ _ _ _ _.
  Next Obligation. unfold dual_le in *; eauto. Qed.
  Next Obligation. unfold dual_le in *; eauto using cola_trans. Qed.
  Next Obligation. unfold dual_le in *; eauto using cola_glb_lower. Qed.
  Next Obligation. unfold dual_le in *; eauto using cola_glb_greatest. Qed.
  Next Obligation. unfold dual_le in *; eauto using cola_lub_upper. Qed.
  Next Obligation. unfold dual_le in *; eauto using cola_lub_least. Qed.
  Next Obligation. unfold dual_le in *; eauto using cola_meet_lower_l. Qed.
  Next Obligation. unfold dual_le in *; eauto using cola_meet_lower_r. Qed.
  Next Obligation. unfold dual_le in *; eauto using cola_meet_greatest. Qed.
  Next Obligation. unfold dual_le in *; eauto using cola_join_upper_l. Qed.
  Next Obligation. unfold dual_le in *; eauto using cola_join_upper_r. Qed.
  Next Obligation. unfold dual_le in *; eauto using cola_join_least. Qed.

End cola_dual.


(** Enabling support for rewriting.
*)

Add Parametric Relation (O : cola) : O ((@cola_le O))
  reflexivity proved by (@cola_refl O)
  transitivity proved by (@cola_trans O)
as cola_le_setoid.

Add Parametric Morphism (O : cola) : (@cola_lub O) with
  signature (@cola_le _) ==> (@cola_le O) 
as cola_lub_mon.
Proof. intros; eapply cola_lub_least; intros; eapply cola_lub_upper; apply H; auto. Qed.

Definition cola_eq (O : cola) (a b : O) := (a <= b /\ b <= a).
Infix "==" := cola_eq (at level 50, no associativity) : cola_scope.

Lemma cola_eq_refl O : reflexive _ (@cola_eq O).
Proof. repeat red; eauto. Qed.
Hint Immediate cola_eq_refl.

Lemma cola_eq_symm O : symmetric _ (@cola_eq O).
Proof. repeat red; split; destruct H; eauto. Qed.

Lemma cola_eq_trans O : transitive _ (@cola_eq O).
Proof. red; intros; split; destruct H, H0; [rewrite H | rewrite H2]; auto. Qed.

Add Parametric Relation (O : cola) : O ((@cola_eq O))
  reflexivity proved by (@cola_eq_refl O)
  symmetry proved by (@cola_eq_symm O)
  transitivity proved by (@cola_eq_trans O)
as cola_eq_setoid.

Lemma cola_glb_mon: forall (O : cola) (x y : O -> Prop),
  y <= x -> cola_glb x <= cola_glb y.
Proof. 
  intros; eapply cola_glb_greatest; intros; eapply cola_glb_lower; apply H; auto. 
Qed.


(** Least and greatest fixed points.
*)

Definition monotone (O1 O2 : cola) (f : cola_carrier O1 -> cola_carrier O2) 
  := forall x y (LE: x <= y), f x <= f y.
Hint Unfold monotone.

Section Tarski.

  Variable O : cola.
  Variable f : cola_carrier O -> cola_carrier O.
  Hypothesis MON : monotone f. 

  Definition fixp := cola_glb (fun x => f x <= x).

  Lemma le_fixp : f fixp <= fixp.
  Proof.
    eapply cola_glb_greatest.
    intros; eapply (cola_trans), H.
    apply MON, cola_glb_lower; auto.
  Qed.

  Lemma fixp_le : fixp <= f fixp.
  Proof.
    eapply cola_glb_lower, MON, le_fixp.
  Qed.

  Lemma fixp_ind : forall u (IH: f u <= u), fixp <= u.
  Proof.
    intros; apply cola_glb_lower; eauto.
  Qed.

  Definition cofixp := cola_lub (fun x => x <= f x).

  Lemma cofixp_le : cofixp <= f cofixp.
  Proof.
    eapply cola_lub_least. 
    intros; eapply cola_trans, MON, cola_lub_upper; eauto.
  Qed.

  Lemma le_cofixp : f cofixp <= cofixp.
  Proof.
    eapply cola_lub_upper, MON, cofixp_le.
  Qed.

  Lemma cofixp_coind : forall u (IH: u <= f u), u <= cofixp.
  Proof.
    intros; apply cola_lub_upper; eauto.
  Qed.

End Tarski.

Lemma fixp_mon :
  forall O (f : cola_carrier O -> cola_carrier O) g,
    f <= g ->
    fixp f <= fixp g.
Proof.
  intros; eapply cola_glb_greatest; intros.
  eapply cola_glb_lower, cola_trans, H0; eauto.
Qed.

Lemma cofixp_mon :
  forall O (f : cola_carrier O -> cola_carrier O) g,
    f <= g ->
    cofixp f <= cofixp g.
Proof.
  intros; eapply cola_lub_least; intros.
  eapply cola_lub_upper, cola_trans; [apply H0|]; eauto.
Qed.
